from  .query_components import Bool, Must, Should, Filter, MustNot, Boost, FieldBoost, \
                              Match, NormalizedMatch, Range, ScoredTerm, Term, Terms

from .utils import QueryParserValidator

query_name_to_component = {
  'bool': Bool,
  'must': Must,
  'should': Should,
  'filter': Filter,
  'must_not': MustNot,
  'boost': Boost,
  'field_boost': FieldBoost,
  'match': Match,
  'normalized_match': NormalizedMatch,
  'range': Range,
  'scored_term': ScoredTerm,
  'term': Term,
  'terms': Terms
}


def parse_query(query, source='query'):
  ''' query (dict) => (QueryComponent) '''
  QueryParserValidator.is_single_value_dict(query, source=source)

  query_root = [k for k in query][0]

  QueryParserValidator.is_valid_op(query_root, query_name_to_component, source=source)

  root_body = query[query_root]

  return query_name_to_component[query_root].from_json(root_body, parse_query)



import math

from .query_component import QueryComponent
from .term import Term
from ._or import Or
from ..query_result import EmptyQR, IdsScoreQR, IdsScoresQR
from ..low_level import sorted_multi_sum_scores
from ..utils import QueryParserValidator, QueryValidatorValidator

class _MatchHelper(QueryComponent):
  def __init__(self, field, value):
    super().__init__('_match_helpers')
    self.field = field
    self.value = value
  
  def execute(self, index, score=True):
    ids = index.get_postings(self.field, self.value).getIds()
    score = 1 / math.log(2 + ids.size)
    return IdsScoreQR(ids, score)

class Match(QueryComponent):
  op_name = 'match'
  def __init__(self, field, value, normalize=False):
    super().__init__('Match')
    self.field = field
    self.value = value
    self.normalize = normalize
  
  def validate(self, index, score=True):
    QueryValidatorValidator.is_str(self.field, source=f'{self.name}.field')
    QueryValidatorValidator.is_str(self.value, source=f'{self.name}.value')

    field_type = index.mapping.get_type(self.field)

    QueryValidatorValidator.is_not_none(field_type, source=f'{self.name}.field -> type')
    QueryValidatorValidator.is_valid_type(field_type, {'text', 'keyword'}, source=f'{self.name}.field -> type')
    return True

  def execute(self, index, score=True):
    tokens = index.analyze(self.field, self.value, return_text=False)

    if len(tokens):
      candidate_list = [_MatchHelper(self.field, token).execute(index, score=True) for token in tokens]
      ids_list = [e.getIds() for e in candidate_list]
      scores_list = [e.getScores() for e in candidate_list]
      ids, scores = sorted_multi_sum_scores(ids_list, scores_list, assume_sorted=True)
      if self.normalize:
        max_possible_score = 0
        for e in candidate_list:
          if e.getScores().size:
            max_possible_score += e.getScores()[0]
        if max_possible_score != 0:
          scores /= max_possible_score
      return IdsScoresQR(ids, scores)
    else:
      return EmptyQR()
  
  def required_rows(self, index, score=True):
    tokens = index.analyze(self.field, self.value, return_text=False)
    req_rows = set()
    for token in tokens:
      req_rows.add((index.get_row_key(self.field, token), self.field))
    return req_rows
  
  @classmethod
  def from_json(cls, query, parse_query_fn):
    QueryParserValidator.is_single_value_dict(query, source=cls.op_name)

    field = [k for k in query][0]
    value = query[field]

    QueryParserValidator.is_str(value, source=f'{cls.op_name}.{field}')

    return cls(field, value)

class NormalizedMatch(Match):
  op_name = 'normalized_match'
  def __init__(self, field, value):
    super().__init__(field, value, normalize=True)
    self.name = 'NormalizedMatch'


from .query_component import QueryComponent
from .must import Must
from .should import Should
from ._filter import Filter
from .must_not import MustNot
from ..low_level import sorted_in1d, combine_must_should
from ..query_result import IdsQR, IdsScoresQR
from ..utils import QueryParserValidator

class Bool(QueryComponent):
  op_name = 'bool'
  def __init__(self, must=None, should=None, _filter=None, must_not=None):
    super().__init__('Bool')
    self.must = must
    self.should = should
    self._filter = _filter
    self.must_not = must_not
  
  def validate(self, index, score=True):
    if self.must is None and self._filter is None and self.should is None:
      raise Exception('bool must specify at least one of [must, should, _filter]')
    
    if self.must is not None:
      if not isinstance(self.must, Must):
        raise Exception('bool.must is a non-Must')
      self.must.validate(index, score=True)

    if self.should is not None:
      if not isinstance(self.should, Should):
        raise Exception('bool.should is a non-Should')
      self.should.validate(index, score=True)

    if self._filter is not None:
      if not isinstance(self._filter, Filter):
        raise Exception('bool._filter is a non-Filter')
      self._filter.validate(index, score=False)

    if self.must_not is not None:
      if not isinstance(self.must_not, MustNot):
        raise Exception('bool.must_not is a non-MustNot')
      self.must_not.validate(index, score=False)

    return True
    

  def execute(self, index, score=True):
    has_filter = self._filter is not None
    has_must_not = self.must_not is not None
    has_must = self.must is not None
    has_should = self.should is not None

    has_any_filter = has_filter or has_must_not
    has_any_score = has_must or has_should

    has_filter_and_must_not = has_filter and has_must_not
    has_must_and_should = has_must and has_should

    if has_any_filter:
      combined_filter_ids = None
      combined_filter_invert = False
      if has_filter:
        _filter = self._filter.execute(index, score=False)
        if has_must_not:
          # if it has both, we can combine them to a single ids list
          _must_not = self.must_not.execute(index, score=False)

          # using sorted_in1d with invert=True returns the ids in the first arg that are not in the second
          combined_filter_ids = _filter.getIds()[sorted_in1d(_filter.getIds(), _must_not.getIds(), invert=True, assume_sorted=True)]
          combined_filter_invert = False
        else:
          combined_filter_ids = _filter.getIds()
          combined_filter_invert = False
      else:
        combined_filter_ids = self.must_not.execute(index, score=False).getIds()
        combined_filter_invert = True

      if has_must_and_should:
        _should = self.should.execute(index, score=True)
        _must = self.must.execute(index, score=True)

        filtered_must_mask = sorted_in1d(_must.getIds(), combined_filter_ids, invert=combined_filter_invert)
        _filtered_must_qr = IdsScoresQR(_must.getIds()[filtered_must_mask], _must.getScores()[filtered_must_mask])
        fm_s_ids, fm_s_scores = combine_must_should(_filtered_must_qr.getIds(), _filtered_must_qr.getScores(), _should.getIds(), _should.getScores(), assume_sorted=True)
        return IdsScoresQR(fm_s_ids, fm_s_scores)
      elif has_must:
        _must = self.must.execute(index, score=True)

        filtered_must_mask = sorted_in1d(_must.getIds(), combined_filter_ids, invert=combined_filter_invert, assume_sorted=True)
        _filtered_must_qr = IdsScoresQR(_must.getIds()[filtered_must_mask], _must.getScores()[filtered_must_mask])
        return _filtered_must_qr
      elif has_should:
        _should = self.should.execute(index, score=True)

        filtered_should_mask = sorted_in1d(_should.getIds(), combined_filter_ids, invert=combined_filter_invert, assume_sorted=True)
        _filtered_should_qr = IdsScoresQR(_should.getIds()[filtered_should_mask], _should.getScores()[filtered_should_mask])
        return _filtered_should_qr
      else:
        return IdsQR(combined_filter_ids)  
    else:
      # there are not filter / must_not
      # if there is must and should we combine, otherwise, return the one present
      if has_must_and_should:
        # combine must and should
        _should = self.should.execute(index, score=True)
        _must = self.must.execute(index, score=True)
        fm_s_ids, fm_s_scores = combine_must_should(_must.getIds(), _must.getScores(), _should.getIds(), _should.getScores(), assume_sorted=True)
        return IdsScoresQR(fm_s_ids, fm_s_scores)
      elif has_must:
        _must = self.must.execute(index, score=True)
        return _must
      else:
        _should = self.should.execute(index, score=True)
        return _should
  
  def required_rows(self, index, score=True):
    req_rows = set()

    if self.must is not None:
      rr = self.must.required_rows(index, score=score)
      req_rows = req_rows.union(rr)
    
    if self.should is not None:
      rr = self.should.required_rows(index, score=score)
      req_rows = req_rows.union(rr)
    
    if self._filter is not None:
      rr = self._filter.required_rows(index, score=score)
      req_rows = req_rows.union(rr)
    
    if self.must_not is not None:
      rr = self.must_not.required_rows(index, score=score)
      req_rows = req_rows.union(rr)

    return req_rows

  def from_json(query, parse_query_fn):
    QueryParserValidator.is_dict(query, source='bool')
    QueryParserValidator.is_not_empty(query, source='bool')
    valid_ops = {'must', 'should', 'must_not', 'filter'}
    for key in query:
      QueryParserValidator.is_valid_op(key, valid_ops, source='bool')

    _must = query.get('must')
    _should = query.get('should')
    _filter = query.get('filter')
    _must_not = query.get('must_not')

    if _must is not None:
      _must = parse_query_fn({'must': _must})
    if _should is not None:
      _should = parse_query_fn({'should': _should})
    if _filter is not None:
      _filter = parse_query_fn({'filter': _filter})
    if _must_not is not None:
      _must_not = parse_query_fn({'must_not': _must_not})

    return Bool(must=_must, should=_should, _filter=_filter, must_not=_must_not)


import time

from multiprocessing.pool import ThreadPool
from .index import Index

class ShardedIndex(Index):
  '''
  An index type that fans out incoming queries to each of the shards. 
  '''
  def __init__(self, name,
    project_id=None,
    gcs_bucket_name=None,
    schema_blob_name=None,
    schema=None,
    shards=[],
    max_threads=4):
    super().__init__(name, 
      project_id=project_id,
      gcs_bucket_name=gcs_bucket_name,
      schema_blob_name=schema_blob_name,
      schema=schema)
    self.shards = shards
    self.max_threads = max_threads
    self.pool = ThreadPool(processes=min(len(shards), self.max_threads))
  
  def query_shard(self, shard_params_tuple):
    shard_num, query_info = shard_params_tuple

    query = query_info.get('query')

    return self.shards[shard_num].query(query,
      limit=query_info.get('limit'),
      offset=query_info.get('offset'),
      min_score=query_info.get('min_score'),
      return_scores=query_info.get('return_scores'),
      return_doc_ids=query_info.get('return_doc_ids'),
      return_documents=query_info.get('return_documents'),
      prefetch=query_info.get('prefetch')
    )

  
  def query(self, query,
      limit=10,
      offset=0,
      min_score=0,
      return_scores=True,
      return_doc_ids=True,
      return_documents=False,
      prefetch=True):
    
    query_info = {
      'query': query,
      'limit': limit + offset,
      'offset': 0,
      'min_score': 0,
      'return_scores': True,
      'return_doc_ids': True,
      'return_documents': return_documents,
      'prefetch': prefetch
    }

    sharded_queries = [(shard, query_info) for shard in range(len(self.shards))]

    query_results = self.pool.map(self.query_shard, sharded_queries)

    response = {
      'parsing_took': [qr.get('parsing_took') for qr in query_results],
      'validation_took': [qr.get('validation_took') for qr in query_results],
      'prefetch_took': [qr.get('prefetch_took') for qr in query_results],
      'query_took': [qr.get('query_took') for qr in query_results],
      'response_took_total': [qr.get('response_took_total') for qr in query_results]
    }
    shards_ids = [qr.get('docs', []) for qr in query_results]
    shards_scores = [qr.get('scores', []) for qr in query_results]
    ids = []
    scores = []
    shard_pointers = [0 for _ in query_results]
    for i in range(limit + offset):
      curr_max = None
      shard_of_max = -1
      for j in range(len(shard_pointers)):
        if len(shards_scores[j]) > shard_pointers[j] and (curr_max is None or shards_scores[j][shard_pointers[j]] > curr_max):
          curr_max = shards_scores[j][shard_pointers[j]]
          shard_of_max = j
      if shard_of_max != -1:
        ids.append(shards_ids[shard_of_max][shard_pointers[shard_of_max]])
        scores.append(shards_scores[shard_of_max][shard_pointers[shard_of_max]])
        shard_pointers[shard_of_max] += 1
    
    response['scores'] = scores[offset:limit+offset]
    response['ids'] = ids[offset:limit+offset]
    return response






import json
import time

from google.cloud import storage

from ..utils import get_blob_string_from_bucket, row_key_formatters
from ..mapping import Mapping
from ..analyzers import analyze
from ..query_result import EmptyQR
from ..query_components import QueryComponent
from ..query_parser import parse_query

from cachetools import cached, LRUCache, TTLCache


class Index(object):
  """
  A logical representation of an HSE index.
  Maintains in-memory caches for performance.
  """
  def __init__(self, name,
    project_id=None,
    gcs_bucket_name=None,
    schema_blob_name=None,
    schema=None):
    self.project_id = project_id
    self.storage_client = storage.Client(self.project_id)
    self.bucket = self.storage_client.bucket(gcs_bucket_name)

    self.load_schema(schema, schema_blob_name)
    self.mapping = Mapping(self.schema.get('mapping', {}))

    self.ttl_for_ttled_postings = int(self.schema.get('ttl', 60*5)) # 5 minutes default
    self.ttled_fields = set(self.schema.get('ttled_fields', []))
    
    self.lru_cache = LRUCache(10000)
    self.ttl_cache = TTLCache(10000, self.ttl_for_ttled_postings)
  
  def load_schema(self, schema, schema_blob_name):
    if schema is not None:
      if type(schema) is not dict:
        raise Exception('When schema is specified, it must be of type dict')
      self.schema = schema
    else:
      if schema_blob_name is None:
        raise Exception('Must specify either schema or schema_blob_name')
      else:
        if type(schema_blob_name) is not str:
          raise Exception('schema_blob_name must be type str')
        self.schema = json.loads(get_blob_string_from_bucket(self.bucket, schema_blob_name))

  def fetch_postings(self, row_key, field_type):
    return EmptyQR()
  
  def get_row_key(self, field, value):
    field_type = self.mapping.get_type(field)
    row_key = row_key_formatters[field_type](field, value)
    return row_key
  
  def batch_fetch_postings(self, row_keys_and_fields):
    pass

  def get_postings(self, field, value):
    row_key = self.get_row_key(field, value)
    field_type = self.mapping.get_type(field)

    cached_value = None

    has_ttl = field in self.ttled_fields

    if has_ttl:
      cached_value = self.ttl_cache.get(row_key)
    else:
      cached_value = self.lru_cache.get(row_key)
    
    if cached_value is not None:
      return cached_value
    else:
      fetched_qr = self.fetch_postings(row_key, field_type)
      if has_ttl:
        self.ttl_cache[row_key] = fetched_qr
      else:
        self.lru_cache[row_key] = fetched_qr
      return fetched_qr
      return fetched_qr
  
  def get_documents(self, doc_ids):
    return []
  
  def has_row_key_cached(self, row_key, field=None):
    if field is not None:
      if field in self.ttled_fields:
        return row_key in self.ttl_cache
      else:
        return row_key in self.lru_cache
    else:
      return row_key in self.lru_cache or row_key in self.ttl_cache
  
  def analyze(self, field, text, return_text=False):
    assert type(text) is str
    analyzer_type = self.mapping.get_analyzer(field)
    return analyze(text, analyzer=analyzer_type, return_text=return_text)

  def query(self, query,
      limit=10,
      offset=0,
      min_score=0,
      return_scores=True,
      return_doc_ids=True,
      return_documents=False,
      prefetch=True):
    print(json.dumps(query) + '\n')
    time_start = time.time()

    time_query_parse_start = time.time()

    try:
      query_object = None
      if isinstance(query, QueryComponent):
        query_object = query
      elif isinstance(query, dict):
        query_object = parse_query(query)
      time_query_parse_done = time.time()
      
      time_validate_start = time.time()
      query_object.validate(self, score=True)
      time_validate_done = time.time()

      time_prefetch_start = time.time()
      if prefetch:
        required_rows = query_object.required_rows(self, score=True)
        self.batch_fetch_postings(required_rows)
      time_prefetch_done = time.time()
      
      time_execution_start = time.time()
      ids, scores = query_object.execute(self, score=True).getTopN(limit + offset)
      time_execution_done = time.time()

      start = min(offset, ids.size)
      end = min(offset + limit, ids.size)

      doc_ids = [int(_id) for _id in ids[start:end]]

      response = {}
      if return_scores:
        response['scores'] = [float(score) for score in scores[start:end]]
      if return_doc_ids:
        response['docs'] = doc_ids
      if return_documents:
        response['hits'] = self.get_documents(doc_ids)

      time_finish = time.time()

      response['parsing_took'] = time_query_parse_done - time_query_parse_start
      response['validation_took'] = time_validate_done - time_validate_start
      response['prefetch_took'] = time_prefetch_done - time_prefetch_start
      response['query_took'] = time_execution_done - time_execution_start
      response['response_took_total'] = time_finish - time_start

      return response
    except Exception as e:
      return {
        'error': str(e)
      }



from .query_component import QueryComponent
from ..query_result import IdsScoreQR
from ..utils import QueryParserValidator
import numpy as np

class Range(QueryComponent):
  op_name = 'range'
  def __init__(self, field, gt=None, gte=None, lt=None, lte=None):
    super().__init__('range')
    self.field = field
    self.gt = gt
    self.gte = gt
    self.lt = lt
    self.lte = lte

    self.ops = {
        'gt': np.greater,
        'gte': np.greater_equal,
        'lt': np.less,
        'lte': np.less_equal
    }

  def validate(self, index, score=False):
    if self.gt is None and self.gte is None and self.lt is None and self.lte is None:
      raise Exception('%s must have at least one of [gt, gte, lt, lte]' % self.name)
    
    if type(self.field) is not str:
      raise Exception('%s.field must be a string' % self.name)

    field_type = index.mapping.get_type(self.field)
    if field_type is None:
      raise Exception('%s.field %s - type not in mapping' % (self.name, self.field))

    if field_type not in {'integer', 'float'}:
      raise Exception('%s.field %s - cannot perform range on type (%s)' % (self.name, self.field, field_type))

    if self.gt is not None and not isinstance(self.gt, (float, int)):
      raise Exception('range.gt must be number when specified')

    if self.gte is not None and not isinstance(self.gte, (float, int)):
      raise Exception('range.gte must be number when specified')

    if self.lt is not None and not isinstance(self.lt, (float, int)):
      raise Exception('range.lt must be number when specified')

    if self.lte is not None and not isinstance(self.lte, (float, int)):
      raise Exception('range.lte must be number when specified')

    return True

  def execute(self, index, score=False):
    values = index.get_postings(self.field, None)
    range_mask = None

    bounds = {
      'gt': self.gt,
      'gte': self.gte,
      'lt': self.lt,
      'lte': self.lte,
    }

    for bound in bounds:
      if bounds[bound] is not None:
        op = self.ops.get(bound)
        mask = op(values.scores, bounds[bound])
        if range_mask is None:
          range_mask = mask
        else:
          range_mask = np.logical_and(range_mask, mask)
    return IdsScoreQR(values.ids[range_mask], 1)
  
  def required_rows(self, index, score=True):
    req_rows = set()
    req_rows.add((index.get_row_key(self.field, None), self.field))
    return req_rows
  
  @classmethod
  def from_json(cls, query, parse_query_fn):
    QueryParserValidator.is_single_value_dict(query, source=cls.op_name)
    field = [k for k in query][0]
    bounds = query[field]

    QueryParserValidator.is_dict(bounds, source=f'{cls.op_name}.{field}')
    QueryParserValidator.is_not_empty(bounds, source=f'{cls.op_name}.{field}')

    allowed_bounds = {'gt', 'gte', 'lt', 'lte'}
    for bound in bounds:
      QueryParserValidator.is_valid_op(bound, allowed_bounds, source=f'{cls.op_name}.{field}.{bound}')
      QueryParserValidator.is_number(bounds[bound], source=f'{cls.op_name}.{field}.{bound}')
    
    _gt = bounds.get('gt')
    _gte = bounds.get('gte')
    _lt = bounds.get('lt')
    _lte = bounds.get('lte')

    return cls(field, gt=_gt, gte=_gte, lt=_lt, lte=_lte)





    






import base64

from google.cloud import storage

def get_blob_string_from_bucket(bucket, blob_name):
  blob = bucket.blob(blob_name)
  return blob.download_as_string().decode('utf-8')

def set_blob_string_from_bucket(bucket, blob_name, content):
  blob = bucket.blob(blob_name)
  blob.upload_from_string(content)
  return True

def get_type_of_field_from_mapping(field, mapping):
    assert type(field) is str
    field_type = None
    spec = {}
    spec.update(mapping)
    for f in field.split('.'):
        spec = spec.get('properties', spec)
        spec = spec.get(f, {})
    field_type = spec.get('type')
    return field_type

def keyword_row_key_formatter(field, value):
    encoded_value = base64.b64encode(value.encode()).decode('utf-8')
    row_key = '%s|%s' % (field, encoded_value)
    return row_key

def integer_row_key_formatter(field, value):
    return field

def float_row_key_formatter(field, value):
    return field

def boolean_row_key_formatter(field, value):
    if type(value) is str:
        return '%s|%s' % (field, value)
    else:
        return '%s|%s' % (field, 'true' if value else 'false')

def doc_row_key_formatter(doc_id, _):
    return 'doc|%s' % (doc_id)

row_key_formatters = {
    'keyword': keyword_row_key_formatter,
    'text': keyword_row_key_formatter,
    'integer': integer_row_key_formatter,
    'float': float_row_key_formatter,
    'boolean': boolean_row_key_formatter,
    'scored_keyword': keyword_row_key_formatter,
    'doc': doc_row_key_formatter
}

class Validator:
  validation_type = ''

  @classmethod
  def is_single_value_dict(cls, value, source=''):
    if type(value) is not dict:
      raise Exception('%s[%s] Error: must be dict' % (cls.validation_type, source))

    if len(value) != 1:
      raise Exception('%s[%s] Error: dict must have a single root.' % (cls.validation_type, source))
  
  @classmethod
  def is_dict(cls, value, source=''):
    if type(value) is not dict:
      raise Exception('%s[%s] Error: value must be dict' % (cls.validation_type, source))
  
  @classmethod
  def is_str(cls, value, source=''):
    if type(value) is not str:
      raise Exception('%s[%s] Error: value must be str' % (cls.validation_type, source))
  
  @classmethod
  def is_number(cls, value, source=''):
    if type(value) is not int and type(value) is not float:
      raise Exception('%s[%s] Error: value must be a number' % (cls.validation_type, source))
  
  @classmethod
  def is_list(cls, value, source=''):
    if type(value) is not list:
      raise Exception('%s[%s] Error: value must be list' % (cls.validation_type, source))
  
  @classmethod
  def is_type(cls, value, _type, type_name, source=''):
    if type(value) is not _type:
      raise Exception('%s[%s] Error: value must be %s' % (cls.validation_type, source, type_name))
  
  @classmethod
  def is_instance(cls, value, _type, type_name, source=''):
    if not isinstance(value, _type):
      raise Exception('%s[%s] Error: value must be instance of %s' % (cls.validation_type, source, type_name))
  
  @classmethod
  def is_not_empty(cls, value, source=''):
    if len(value) == 0:
      raise Exception('%s[%s] Error: value must be non-empty' % (cls.validation_type, source))
  
  @classmethod
  def is_not_none(cls, value, source=''):
    if value is None:
      raise Exception('%s[%s] Error: value is None' % (cls.validation_type, source))

  @classmethod
  def is_nonempty_list(cls, value, source=''):
    if type(value) is not list:
      raise Exception('%s[%s] Error: value must be list' % (cls.validation_type, source))
    if len(value) == 0:
      raise Exception('%s[%s] Error: value must be non-empty list' % (cls.validation_type, source))
  
  @classmethod
  def is_list_of_str(cls, value, source=''):
    cls.is_list(value, source=source)

    for v in value:
      cls.is_str(v, source=source + '[:]')
    

class QueryParserValidator(Validator):
  validation_type = 'QueryParser'

  @classmethod
  def is_valid_op(cls, value, valid_ops, source=''):
    if value not in valid_ops:
      raise Exception('%s[%s] Error: unknown query operation (%s)' % (cls.validation_type, source, value))
    
class QueryValidatorValidator(Validator):
  validation_type = 'QueryValidator'

  @classmethod
  def is_valid_type(cls, value, valid_types, source=''):
    if value not in valid_types:
      raise Exception('%s[%s] Error: invalid type (%s)' % (cls.validation_type, source, value))



cimport cython
import numpy as np
import heapq
cimport numpy as np

DTYPE = np.float
ctypedef np.float_t DTYPE_t

cdef class IdsScores:
    cdef int[:] ids
    cdef float[:] scores
    def __init__(self, int[:] ids, float[:] scores):
        self.ids = ids
        self.scores = scores

cdef class Ids:
    cdef int[:] ids
    def __init__(self, int[:] ids):
        self.ids = ids

@cython.boundscheck(False)
def is_sorted(int[:] a):
    cdef int a_last = -2147483648
    cdef int al = a.size

    for ai in range(al):
        if a[ai] < a_last:
            return False
        else:
            a_last = a[ai]

    return True

@cython.boundscheck(False)
def sorted_intersect1d(int[:] a, int[:] b, assume_sorted=False):
    if not assume_sorted:
        assert is_sorted(a) and is_sorted(b)
    cdef int[:] ab = np.empty(min(a.size, b.size), dtype=np.int32)

    cdef int bi = 0

    cdef int al = a.size
    cdef int bl = b.size
    cdef int n_int = 0

    for ai in range(al):
        while a[ai] > b[bi] and bi < bl:
            bi += 1
        if bi == bl:
            break
        else:
            if a[ai] == b[bi]:
                # if they're the same, we add to the intersection
                # otherwise, b is greater, so we need to advance ai
                ab[n_int] = a[ai]
                bi += 1
                n_int += 1
    return np.array(ab[0:n_int])

@cython.boundscheck(False)
def sorted_in1d(int[:] a, int[:] b, assume_sorted=False, invert=False):
    if not assume_sorted:
        assert is_sorted(a) and is_sorted(b)
    cdef int[:] ab = np.ones(a.size, dtype=np.int32) if invert else np.zeros(a.size, dtype=np.int32)

    cdef int bi = 0

    cdef int al = a.size
    cdef int bl = b.size

    for ai in range(al):
        while a[ai] > b[bi] and bi < bl:
            bi += 1
        if bi == bl:
            break
        else:
            if a[ai] == b[bi]:
                # if they're the same, we add to the intersection
                # otherwise, b is greater, so we need to advance ai
                ab[ai] = 0 if invert else 1
                bi += 1
    return np.array(ab, dtype=np.bool)


@cython.boundscheck(False)
cdef int[:] sorted_union1d_internal(int[:] a, int[:] b, assume_sorted=False):
    if not assume_sorted:
        assert is_sorted(a) and is_sorted(b)
    cdef int[:] ab = np.empty(a.size + b.size, dtype=np.int32)

    cdef int bi = 0

    cdef int al = a.size
    cdef int bl = b.size
    cdef int n_int = 0

    for ai in range(al):
        while a[ai] > b[bi] and bi < bl:
            ab[n_int] = b[bi]
            bi += 1
            n_int += 1 
        ab[n_int] = a[ai]
        n_int += 1
        if a[ai] == b[bi]:
            bi += 1
    while bi < bl:
        ab[n_int] = b[bi]
        bi += 1
        n_int += 1
         
    return ab[0:n_int]

@cython.boundscheck(False)
def sorted_union1d(int[:] a, int[:] b, assume_sorted=False):
    return np.array(sorted_union1d_internal(a, b, assume_sorted=assume_sorted), dtype=np.int32)

@cython.boundscheck(False)
cdef int[:] sorted_multi_union1d_internal(list ids_list, assume_sorted=False):
    cdef int l = len(ids_list)
    cdef int split_i = 0
    cdef int[:] a_ids, b_ids
    if l == 1:
        return ids_list[0]
    elif l == 2:
        return sorted_union1d_internal(ids_list[0], ids_list[1], assume_sorted=assume_sorted)
    else:
        # we have more than two, so we need to divide
        split_i = l // 2
        a_ids = sorted_multi_union1d_internal(ids_list[:split_i], assume_sorted=assume_sorted)
        b_ids = sorted_multi_union1d_internal(ids_list[split_i:], assume_sorted=assume_sorted)

        return sorted_union1d_internal(a_ids, b_ids, assume_sorted=assume_sorted)

@cython.boundscheck(False)
def sorted_multi_union1d(list ids_list, assume_sorted=False):
    cdef int[:] ids = sorted_multi_union1d_internal(ids_list, assume_sorted=assume_sorted)
    return np.array(ids, dtype=np.int32)

@cython.boundscheck(False)
cdef IdsScores sorted_sum_scores_internal(int[:] a, float[:] af, int[:] b, float[:] bf, assume_sorted=False):
    if not assume_sorted:
        assert is_sorted(a) and is_sorted(b)
    cdef int[:] ab = np.empty(a.size + b.size, dtype=np.int32)
    cdef float[:] abf = np.empty(a.size + b.size, dtype=np.float32)

    cdef int bi = 0

    cdef int al = a.size
    cdef int bl = b.size
    cdef int n_int = 0

    for ai in range(al):
        while a[ai] > b[bi] and bi < bl:
            ab[n_int] = b[bi]
            abf[n_int] = bf[bi]
            bi += 1
            n_int += 1 
        ab[n_int] = a[ai]
        if a[ai] == b[bi]:
            abf[n_int] = af[ai] + bf[bi]
            bi += 1
        else:
            abf[n_int] = af[ai]
        n_int += 1
        
    while bi < bl:
        ab[n_int] = b[bi]
        abf[n_int] = bf[bi]
        bi += 1
        n_int += 1
         
    return IdsScores(ab[0:n_int], abf[0:n_int])

@cython.boundscheck(False)
def sorted_sum_scores(int[:] a, float[:] af, int[:] b, float[:] bf, assume_sorted=False):
    cdef IdsScores ids_scores = sorted_sum_scores_internal(a, af, b, bf, assume_sorted=assume_sorted)
    return (np.array(ids_scores.ids, dtype=np.int32), np.array(ids_scores.scores, dtype=np.float32))

@cython.boundscheck(False)
cdef IdsScores sorted_multi_sum_scores_internal(list ids_list, list scores_list, assume_sorted=False):
    cdef int l = len(ids_list)
    cdef int split_i = 0
    cdef IdsScores a_ids_scores, b_ids_scores
    if l == 1:
        return IdsScores(ids_list[0], scores_list[0])
    elif l == 2:
        return sorted_sum_scores_internal(ids_list[0], scores_list[0], ids_list[1], scores_list[1], assume_sorted=assume_sorted)
    else:
        # we have more than two, so we need to divide
        split_i = l // 2
        a_ids_scores = sorted_multi_sum_scores_internal(ids_list[:split_i], scores_list[:split_i], assume_sorted=assume_sorted)
        b_ids_scores = sorted_multi_sum_scores_internal(ids_list[split_i:], scores_list[split_i:], assume_sorted=assume_sorted)

        return sorted_sum_scores_internal(a_ids_scores.ids, a_ids_scores.scores, b_ids_scores.ids, b_ids_scores.scores, assume_sorted=assume_sorted)

@cython.boundscheck(False)
def sorted_multi_sum_scores(list ids_list, list scores_list, assume_sorted=False):
    cdef IdsScores ids_scores = sorted_multi_sum_scores_internal(ids_list, scores_list, assume_sorted=assume_sorted)
    return (np.array(ids_scores.ids, dtype=np.int32), np.array(ids_scores.scores, dtype=np.float32))

@cython.boundscheck(False)
cdef IdsScores sorted_sum_scores_must_internal(int[:] a, float[:] af, int[:] b, float[:] bf, assume_sorted=False):
    if not assume_sorted:
        assert is_sorted(a) and is_sorted(b)
    
    cdef int[:] ab = np.empty(min(a.size, b.size), dtype=np.int32)
    cdef float[:] abf = np.empty(min(a.size, b.size), dtype=np.float32)

    cdef int bi = 0

    cdef int al = a.size
    cdef int bl = b.size
    cdef int n_int = 0

    for ai in range(al):
        while a[ai] > b[bi] and bi < bl:
            bi += 1
        if bi == bl:
            break
        else:
            if a[ai] == b[bi]:
                # if they're the same, we add to the intersection
                # otherwise, b is greater, so we need to advance ai
                ab[n_int] = a[ai]
                abf[n_int] = af[ai] + bf[bi]
                bi += 1
                n_int += 1
    return IdsScores(ab[0:n_int], abf[0:n_int])

def sorted_sum_scores_must(int[:] a, float[:] af, int[:] b, float[:] bf, assume_sorted=False):
    cdef IdsScores ids_scores = sorted_sum_scores_must_internal(a, af, b, bf, assume_sorted=assume_sorted)
    return (np.array(ids_scores.ids), np.array(ids_scores.scores))


@cython.boundscheck(False)
cdef IdsScores sorted_multi_sum_scores_must_internal(list ids_list, list scores_list, assume_sorted=False):
    cdef int l = len(ids_list)
    cdef int split_i = 0
    cdef IdsScores a_ids_scores, b_ids_scores
    if l == 1:
        return IdsScores(ids_list[0], scores_list[0])
    elif l == 2:
        return sorted_sum_scores_must_internal(ids_list[0], scores_list[0], ids_list[1], scores_list[1], assume_sorted=assume_sorted)
    else:
        # we have more than two, so we need to divide
        split_i = l // 2
        a_ids_scores = sorted_multi_sum_scores_must_internal(ids_list[:split_i], scores_list[:split_i], assume_sorted=assume_sorted)
        b_ids_scores = sorted_multi_sum_scores_must_internal(ids_list[split_i:], scores_list[split_i:], assume_sorted=assume_sorted)

        return sorted_sum_scores_must_internal(a_ids_scores.ids, a_ids_scores.scores, b_ids_scores.ids, b_ids_scores.scores, assume_sorted=assume_sorted)

@cython.boundscheck(False)
def sorted_multi_sum_scores_must(list ids_list, list scores_list, assume_sorted=False):
    cdef IdsScores ids_scores = sorted_multi_sum_scores_must_internal(ids_list, scores_list, assume_sorted=assume_sorted)
    return (np.array(ids_scores.ids, dtype=np.int32), np.array(ids_scores.scores, dtype=np.float32))

@cython.boundscheck(False)
def combine_must_should(int[:] must_ids, float[:] must_scores, int[:] should_ids, float[:] should_scores, assume_sorted=False):
    if not assume_sorted:
        assert is_sorted(must_ids) and is_sorted(should_ids)
    
    cdef int[:] ab = np.empty(must_ids.size, dtype=np.int32)
    cdef float[:] abf = np.empty(must_ids.size, dtype=np.float32)

    cdef int bi = 0

    cdef int al = must_ids.size
    cdef int bl = should_ids.size

    for ai in range(al):
        ab[ai] = must_ids[ai]
        while must_ids[ai] > should_ids[bi] and bi < bl - 1:
            bi += 1
        if must_ids[ai] == should_ids[bi]:
            # if they're the same, we add their scores
            # otherwise, we just use the must score
            abf[ai] = must_scores[ai] + should_scores[bi]
        else:
            abf[ai] = must_scores[ai]
    return (ab, abf)



from ._or import Or

class Should(Or):
  op_name = 'should'
  def __init__(self, sub_queries):
    super().__init__(sub_queries, name='Should')

  def execute(self, index, score=True):
    return super().execute(index, score=True)



import functools

from .query_component import CompoundQuery
from ..query_result import IdsQR, IdsScoresQR
from ..low_level import sorted_intersect1d, sorted_multi_sum_scores_must

from ..utils import QueryValidatorValidator, QueryParserValidator

'''
{
  'interleave': {
    'queries': [
      {
        ...
      },
      {
        ...
      },
      {
        ...
      }
    ]
  }
}
'''

class Interleave(CompoundQuery):
  op_name = 'interleave'
  def __init__(self, sub_queries, window_size=100):
    super().__init__('Interleave', sub_queries)
    self.window_size = window_size
  
  def validate(self, index, score=True):
    QueryValidatorValidator.is_type(self.window_size, int, 'int', source=f'{self.name}.window_size')
    super().validate(index, score=score)
    return True
  
  def execute(self, index, score=True):
    candidate_list = [q.execute(index, score=score) for q in self.sub_queries]
    if score:
      ids_list = [e.getIds() for e in candidate_list]
      scores_list = [e.getScores() for e in candidate_list]
      ids, scores = sorted_multi_sum_scores_must(ids_list, scores_list, assume_sorted=True)
      return IdsScoresQR(ids, scores)
    else:
      candidate_list.sort(key=lambda x: x.getIds().size)
      return functools.reduce(lambda a, b: IdsQR(sorted_intersect1d(a.getIds(), b.getIds(), assume_sorted=True)), candidate_list)

  @classmethod
  def from_json(cls, query, parse_query_fn):
    QueryParserValidator.is_dict(query, source=cls.op_name)

    window_size = query.get(window_size, 100)
    QueryParserValidator.is_type(window_size, int, 'int', source=f'{cls.op_name}.window_size')

    queries = query.get('queries')
    QueryParserValidator.is_nonempty_list(queries, source=f'{cls.op_name}.queries')

    sub_queries = [parse_query_fn(subq, source=f'{self.op_name}.queries[:]') for subq in queries]

    return cls(sub_queries, window_size=window_size)


from .query_component import CompoundQuery
from ..query_result import IdsQR, IdsScoresQR
from ..low_level import sorted_multi_sum_scores, sorted_multi_union1d

class Or(CompoundQuery):
  op_name = 'or'
  def __init__(self, sub_queries, name='Or'):
    super().__init__(name, sub_queries)
  
  def execute(self, index, score=True):
    candidate_list = [q.execute(index, score=score) for q in self.sub_queries]
    if score:
      ids_list = [e.getIds() for e in candidate_list]
      scores_list = [e.getScores() for e in candidate_list]
      ids, scores = sorted_multi_sum_scores(ids_list, scores_list, assume_sorted=True)
      return IdsScoresQR(ids, scores)
    else:
      ids_list = [e.getIds() for e in candidate_list]
      return IdsQR(sorted_multi_union1d(ids_list, assume_sorted=True))



import numpy as np

from .query_component import QueryComponent
from ..query_result import IdsScoresQR
from ..utils import QueryParserValidator, QueryValidatorValidator

class Boost(QueryComponent):
  op_name = 'boost'
  def __init__(self, query, boost):
    super().__init__('Boost')
    self.query = query
    self.boost = boost
  
  def validate(self, index, score=True):
    QueryValidatorValidator.is_instance(self.query, QueryComponent, 'QueryComponent', source=f'{self.name}.query')
    QueryValidatorValidator.is_number(self.boost, source=f'{self.name}.boost')
    self.query.validate(index, score=score)
    return True

  def execute(self, index, score=True):
    a = self.query.execute(index, score=True)
    if self.boost != 1:
      b = IdsScoresQR(a.ids, np.array(a.scores, dtype=np.float32) * self.boost)
      return b
    else:
      return a
  
  def required_rows(self, index, score=True):
    return self.query.required_rows(index, score=True)

  @classmethod
  def from_json(cls, query, parse_query_fn):
    QueryParserValidator.is_dict(query, source=cls.op_name)

    q = query.get('query')
    boost = query.get('boost')

    QueryParserValidator.is_number(boost, source=f'{cls.op_name}.boost')

    return cls(parse_query_fn(q, source=f'{cls.op_name}.query'), boost)



from .query_component import QueryComponent, CompoundQuery
from ._and import And
from ._or import Or
from ._filter import Filter
from .bool import Bool
from .boost import Boost
from .field_boost import FieldBoost
from .match import Match, NormalizedMatch
from .must_not import MustNot
from .must import Must
from .range import Range
from .scored_term import ScoredTerm
from .should import Should
from .term import Term
from .terms import Terms



class Mapping:
  def __init__(self, mapping):
    self.mapping = mapping
    self.fields_to_types = self.get_field_types_from_mapping(mapping)
    self.fields_to_analyzers = self.get_analyzers_from_mapping(mapping)
  
  def get_type(self, field):
    assert type(field) is str
    return self.fields_to_types.get(field)
  
  def get_analyzer(self, field):
    assert type(field) is str
    return self.fields_to_analyzers.get(field)
  
  def get_field_types_from_mapping(self, mapping):
    ftt = {}
    for field_name in mapping:
      ftype = mapping[field_name].get('type')
      properties = mapping[field_name].get('properties')
      if ftype is not None:
        ftt[field_name] = ftype
      if properties is not None:
        sub_fields = self.get_field_types_from_mapping(properties)
        for sf in sub_fields:
          ftt['%s.%s' % (field_name, sf)] = sub_fields[sf]
    return ftt

  def get_analyzers_from_mapping(self, mapping):
    fta = {}
    for field_name in mapping:
      ftype = mapping[field_name].get('type')
      analyzer_type = mapping[field_name].get('analyzer', 'english')
      properties = mapping[field_name].get('properties')
      if ftype == 'text':
        fta[field_name] = analyzer_type
      if properties is not None:
        sub_fields = self.get_analyzers_from_mapping(properties)
        for sf in sub_fields:
          fta['%s.%s' % (field_name, sf)] = sub_fields[sf]
    return fta


from ._and import And

class MustNot(And):
  '''
  Note, by itself, the must not filter behaves exactly like filter, meaning
  it does not do the 'not' part. In order to get the not of something, this
  op must be used with the Bool query so that it can be included with other
  criteria.
  '''
  op_name = 'must_not'
  def __init__(self, sub_queries):
    super().__init__(sub_queries, name='MustNot')

  def execute(self, index, score=False):
    return super().execute(index, score=False)



import html
import gensim

def tokenize(text):
  s = html.unescape(text)
  s = gensim.utils.simple_preprocess(s)
  return s

stopwords_sets = {
    'english': {"it", "its", "itself", "what", "which", "who", "whom", "this", "that", "these", "those", "am", "is", "are", "was", "were", "be", "been", "being", "have", "has", "had", "having", "do", "does", "did", "doing", "a", "an", "the", "and", "but", "if", "or", "because", "as", "of", "at", "by", "for", "with", "to", "from", "in", "then", "all", "any", "both", "each", "few", "more", "most", "other", "some", "such", "no", "nor", "not", "only", "own", "same", "so", "than", "too", "very", "s", "t", "can", "will", "just", "don", "should", "now"}
}

def analyzeEnglish(text):
  tokens = tokenize(text)
  final_tokens = []
  for token in tokens:
    if token not in stopwords_sets['english']:
      final_tokens.append(token)
  return final_tokens

analyzers = {
  'english': analyzeEnglish
}

def analyze(text, analyzer='english', return_text=False):
  if analyzer is None:
    return [text]
  else:
    tokens = analyzers[analyzer](text)
    if return_text:
      return ' '.join(tokens)
    else:
      return tokens



from ..query_result import EmptyQR
from ..utils import QueryParserValidator, QueryValidatorValidator

class QueryComponent:
  op_name = None
  def __init__(self, name):
    self.name = name
  
  def validate(self, index, score=True):
    try:
      assert type(self.name) is str
    except:
      raise Exception('QueryComponent must have name')
    return True

  def execute(self, index, score=True):
    return EmptyQR()
  
  def required_rows(self, index, score=True):
    return set()
  
  def from_json(value, parse_query_fn):
    return QueryComponent(None)

class CompoundQuery(QueryComponent):
  op_name = None
  def __init__(self, name, sub_queries):
    super().__init__(name)
    self.sub_queries = sub_queries
  
  def validate(self, index, score=True):
    super().validate(index, score=score)
    QueryValidatorValidator.is_nonempty_list(self.sub_queries, source=f'{self.name}[:]')
    
    for subq in self.sub_queries:
      QueryValidatorValidator.is_instance(subq, QueryComponent, 'QueryComponent', source=f'{self.name}[:]')
      subq.validate(index, score=score)
    return True
  
  def required_rows(self, index, score=True):
    req_rows = set()
    for q in self.sub_queries:
      rr = q.required_rows(index, score=score)
      req_rows = req_rows.union(rr)
    return req_rows
  
  @classmethod
  def from_json(cls, query, parse_query_fn):
    QueryParserValidator.is_nonempty_list(query, source=cls.op_name)
    
    for subq in query:
      QueryParserValidator.is_single_value_dict(subq, source=cls.op_name)
    
    sub_queries = [parse_query_fn(subq) for subq in query]

    return cls(sub_queries)



import numpy as np

class QueryResult(object):
  def __init__(self):
    self.ids = None
    self.scores = None
  
  def getIds(self):
    return self.ids
  
  def getScores(self):
    return self.scores
  
  def getTopN(self, n, return_scores=True):
    n_clip = min(n, self.getIds().size)
    
    top_indices = np.flip(np.argpartition(self.getScores(), -n_clip)[-n_clip:])
    top_scores = np.take(self.getScores(), top_indices)
    top_scores_indices = np.flip(np.argsort(top_scores))
    top_scores_sorted = np.take(top_scores, top_scores_indices)
    top_ids_sorted = np.take(np.take(self.getIds(), top_indices), top_scores_indices)
    
    if return_scores:
      return (top_ids_sorted, top_scores_sorted)
    else:
      return (top_ids)

class EmptyQR(QueryResult):
  def __init__(self):
    self.ids = np.array([], dtype=np.int32)
    self.scores = np.array([], dtype=np.float32)

class IdsQR(QueryResult):
  """
  A type of query result consisting of a numpy array of ids (integers)
  ex. This QR is returned by the Term query in filter setting
  """
  def __init__(self, ids):
    super().__init__()
    self.ids = ids
  
  
  def getScores(self):
    if self.scores is None:
      self.scores = np.zeros(self.getIds().size, dtype=np.float32)
    return self.scores
  
  def getTopN(self, n, return_scores=True):
    top_ids = self.getIds()[0:n]
    if return_scores:
      top_scores = np.full(top_ids.size, 0, dtype=np.float32)
      return (top_ids, top_scores)
    else:
      return (top_ids)

class IdsScoreQR(QueryResult):
  """
  A type of query result consisting of a numpy array of ids (integers)
  and a float16 'score' (array([1,2,4]), 1.0). 
  ex. This QR is returned by the Term query in score setting and is how postings are stored
  """
  def __init__(self, ids, score):
    super().__init__()
    self.score = score
    self.ids = ids

  def getScores(self):
    if self.scores is not None:
      return self.scores
    else:
      self.scores = np.full(self.ids.size, self.score, dtype=np.float32)
      return self.scores
  
  def getTopN(self, n, return_scores=True):
    top_ids = self.getIds()[0:n]
    if return_scores:
      top_scores = np.full(top_ids.size, self.score, dtype=np.float32)
      return (top_ids, top_scores)
    else:
      return (top_ids)

class IdsScoresQR(QueryResult):
  """
  A type of query result with np arrays for ids and scores
  """
  def __init__(self, ids, scores):
    self.ids = ids
    self.scores = scores




from ._and import And

class Filter(And):
  op_name = 'filter'
  def __init__(self, sub_queries):
    super().__init__(sub_queries, name='Filter')

  def execute(self, index, score=False):
    return super().execute(index, score=False)



class HTTPIndex(Index):
  '''
  An index type that forwards queries to a searcher located at a specific url.
  Masters might have these to handle sending requests to the right workers.
  '''
  def __init__(self, name,
    project_id=None,
    gcs_bucket_name=None,
    schema_blob_name=None,
    schema=None,
    url=None):
    super().__init__(name, 
      project_id=project_id,
      gcs_bucket_name=gcs_bucket_name,
      schema_blob_name=schema_blob_name,
      schema=schema)
    self.url = url


from .query_component import QueryComponent
from ._or import Or
from .term import Term
from ..utils import QueryParserValidator, QueryValidatorValidator

class Terms(QueryComponent):
  op_name = 'terms'
  def __init__(self, field, values):
    super().__init__('Terms')
    self.field = field
    self.values = values
  
  def validate(self, index, score=True):
    QueryParserValidator.is_str(self.field, source=f'{self.name}.field')
    QueryParserValidator.is_list_of_str(self.values, source=f'{self.name}.{self.field}[:]')

    field_type = index.mapping.get_type(self.field)
    QueryValidatorValidator.is_not_none(field_type, source=f'{self.name}.field -> {self.field}.type')

    QueryValidatorValidator.is_valid_type(field_type, {'text', 'keyword'}, source=f'{self.name}.field -> {self.field} ({field_type})')
    return True
  
  def execute(self, index, score=True):
    return Or([Term(self.field, value) for value in self.values]).execute(index, score=score)
  
  def required_rows(self, index, score=True):
    req_rows = set()
    for value in self.values:
      req_rows.add((index.get_row_key(self.field, value), self.field))
    return req_rows

  @classmethod
  def from_json(cls, query, parse_query_fn):
    QueryParserValidator.is_single_value_dict(query, source=cls.op_name)

    field = [k for k in query][0]

    QueryParserValidator.is_list_of_str(query[field], source=f'{cls.op_name}.{field}')

    value = query[field]
    return Terms(field, value)



from .query_component import QueryComponent
from ..query_result import IdsScoreQR
from ..utils import QueryParserValidator, QueryValidatorValidator

class Term(QueryComponent):
  op_name = 'term'
  def __init__(self, field, value):
    super().__init__('Term')
    self.field = field
    self.value = value
  
  def validate(self, index, score=True):
    QueryValidatorValidator.is_str(self.field, source=self.name)
    QueryValidatorValidator.is_str(self.value, source=self.name)

    field_type = index.mapping.get_type(self.field)

    QueryValidatorValidator.is_not_none(field_type, source=f'{self.name}.field -> type')
    QueryValidatorValidator.is_valid_type(field_type, {'text', 'keyword'}, source=f'{self.name}.field -> type')
    return True

  def execute(self, index, score=True):
    return IdsScoreQR(index.get_postings(self.field, self.value).getIds(), 1)
  
  def required_rows(self, index, score=True):
    req_rows = set()
    req_rows.add((index.get_row_key(self.field, self.value), self.field))
    return req_rows
  
  @classmethod
  def from_json(cls, query, parse_query_fn):
    QueryParserValidator.is_single_value_dict(query, source=cls.op_name)

    field = [k for k in query][0]
    QueryParserValidator.is_single_value_dict(query[field], source=f'{cls.op_name}.{field}')

    value = query[field].get('value')
    QueryParserValidator.is_str(value, source=f'{cls.op_name}.{field}.value')

    return cls(field, value)



import json

import numpy as np

from google.cloud import bigtable
from google.cloud.bigtable import row_filters, row_set

from ..query_result import IdsQR, IdsScoresQR
from .index import Index

def keyword_row_parser(row):
    ids_bytes = row.cells.get('i').get('i'.encode())[0].value
    ids = np.array(np.frombuffer(ids_bytes, dtype=np.int32), dtype=np.int32)
    return IdsQR(ids)

def integer_row_parser(row):
    ids_bytes = row.cells.get('i').get('i'.encode())[0].value
    ids = np.array(np.frombuffer(ids_bytes, dtype=np.int32), dtype=np.int32)
    scores_bytes = row.cells.get('s').get('i'.encode())[0].value
    scores = np.array(np.frombuffer(scores_bytes, dtype=np.float32), dtype=np.float32)
    return IdsScoresQR(ids, scores)

def float_row_parser(row):
    ids_bytes = row.cells.get('i').get('i'.encode())[0].value
    ids = np.array(np.frombuffer(ids_bytes, dtype=np.int32), dtype=np.int32)
    scores_bytes = row.cells.get('s').get('f'.encode())[0].value
    scores = np.array(np.frombuffer(scores_bytes, dtype=np.float32), dtype=np.float32)
    return IdsScoresQR(ids, scores)

def doc_row_parser(row):
    response = row.cells.get('doc').get('doc'.encode())[0].value
    try:
        return json.loads(response)
    except:
        return response

row_parsers = {
    'keyword': keyword_row_parser,
    'text': keyword_row_parser,
    'integer': integer_row_parser,
    'float': float_row_parser,
    'boolean': keyword_row_parser,
    'scored_keyword': float_row_parser,
    'doc': doc_row_parser,
}

class BigTableIndex(Index):
  def __init__(self, name,
    project_id=None,
    gcs_bucket_name=None,
    schema_blob_name=None,
    schema=None,
    instance_id=None,
    table_id=None,
    instance=None,
    table=None):
    super().__init__(name, 
      project_id=project_id,
      gcs_bucket_name=gcs_bucket_name,
      schema_blob_name=schema_blob_name,
      schema=schema)
    
    if table is not None:
      self.table = table
    else:
      if table_id is None:
        raise Exception('table_id or table must be provided')
      if instance is not None:
        self.table = instance.table(table_id)
      else:
        if instance_id is None:
          raise Exception('instance_id or instance must be provided')
        bt_client = bigtable.Client(self.project_id)
        instance = bt_client.instance(instance_id)
        self.table = instance.table(table_id)
    
    self.bt_row_filter = row_filters.CellsColumnLimitFilter(1)
  
  def bt_lookup(self, row_key):
    row = self.table.read_row(row_key.encode('utf-8'), self.bt_row_filter)
    return row

  def fetch_postings(self, row_key, field_type):
    row = self.bt_lookup(row_key)
    fetched_qr = None
    if row is not None:
      fetched_qr = row_parsers[field_type](row)
    else:
      fetched_qr = EmptyQR()
    return fetched_qr

  def batch_fetch_postings(self, row_keys_and_fields):
    row_keys_fields = {
      key: field for key, field in row_keys_and_fields
      if not self.has_row_key_cached(key, field=field)
    }
    row_keys_types = {key: self.mapping.get_type(row_keys_fields[key]) for key in row_keys_fields}

    rs = row_set.RowSet()
    set_size = 0
    for rk in row_keys_fields:
      rs.add_row_key(rk)
      set_size += 1
    if set_size > 0:
      partial_rows_data = self.table.read_rows(row_set=rs, filter_=self.bt_row_filter)
      partial_rows_data.consume_all()

      for rk in row_keys_types:
        row = partial_rows_data.rows.get(rk.encode())
        t = row_keys_types[rk]
        f = row_keys_fields[rk]
        qr = None
        if row is None:
          qr = EmptyQR()
        else:
          qr = row_parsers[t](row)
        
        if f in self.ttled_fields:
          self.ttl_cache[rk] = qr
        else:
          self.lru_cache[rk] = qr
    
  def get_documents(self, doc_ids):
    row_keys = ['doc|%s' % (i) for i in doc_ids]

    documents = []

    rs = row_set.RowSet()
    set_size = 0
    for rk in row_keys:
      rs.add_row_key(rk)
      set_size += 1
    if set_size > 0:
      partial_rows_data = self.table.read_rows(row_set=rs, filter_=self.bt_row_filter)
      partial_rows_data.consume_all()

      for rk in row_keys:
        row = partial_rows_data.rows.get(rk.encode())
        if row is not None:
          doc = row_parsers['doc'](row)
          documents.append(doc)
        else:
          documents.append(None)
    return documents


import functools

from .query_component import CompoundQuery
from ..query_result import IdsQR, IdsScoresQR
from ..low_level import sorted_intersect1d, sorted_multi_sum_scores_must

class And(CompoundQuery):
  op_name = 'and'
  def __init__(self, sub_queries, name='And'):
    super().__init__(name, sub_queries)
  
  def execute(self, index, score=True):
    candidate_list = [q.execute(index, score=score) for q in self.sub_queries]
    if score:
      ids_list = [e.getIds() for e in candidate_list]
      scores_list = [e.getScores() for e in candidate_list]
      ids, scores = sorted_multi_sum_scores_must(ids_list, scores_list, assume_sorted=True)
      return IdsScoresQR(ids, scores)
    else:
      candidate_list.sort(key=lambda x: x.getIds().size)
      return functools.reduce(lambda a, b: IdsQR(sorted_intersect1d(a.getIds(), b.getIds(), assume_sorted=True)), candidate_list)



from ._and import And

class Must(And):
  op_name = 'must'
  def __init__(self, sub_queries):
    super().__init__(sub_queries, name='Must')
  
  def execute(self, index, score=True):
    return super().execute(index, score=True)



from setuptools import setup, Extension
from Cython.Distutils import build_ext
import numpy as np

NAME = "hse-search"
VERSION = "0.0.2"
DESCR = "!"
REQUIRES = [
  'numpy', 
  'cython',
  'google-cloud-storage==1.19.1',
  'google-cloud-bigtable',
  'gensim',
  'cachetools'
]

AUTHOR = "Mark Hudson"
EMAIL = "mark.cd.hudson@gmail.com"

LICENSE = "Apache 2.0"

SRC_DIR = "hse_search"
PACKAGES = [
  SRC_DIR,
  SRC_DIR + ".low_level",
  SRC_DIR + ".query_components",
  SRC_DIR + ".index"
]

ext_1 = Extension(SRC_DIR + ".low_level.wrapped",
                  [SRC_DIR + "/low_level/wrapped.pyx"],
                  libraries=[],
                  include_dirs=[np.get_include()])


EXTENSIONS = [ext_1]

if __name__ == "__main__":
  setup(install_requires=REQUIRES,
    packages=PACKAGES,
    zip_safe=False,
    name=NAME,
    version=VERSION,
    description=DESCR,
    author=AUTHOR,
    author_email=EMAIL,
    license=LICENSE,
    cmdclass={"build_ext": build_ext},
    ext_modules=EXTENSIONS
  )



import numpy as np

from .query_component import QueryComponent
from ..query_result import IdsScoresQR
from ..utils import QueryValidatorValidator, QueryParserValidator

class FieldBoost(QueryComponent):
  op_name = 'field_boost'
  def __init__(self, field, boost):
    super().__init__('FieldBoost')
    self.field = field
    self.boost = boost
  
  def validate(self, index, score=True):
    QueryValidatorValidator.is_str(self.field, source=f'{self.name}.field')
    field_type = index.mapping.get_type(self.field)
    QueryValidatorValidator.is_not_none(field_type, source=f'{self.name}.field -> type')
    QueryValidatorValidator.is_valid_type(field_type, {'integer', 'float'}, source=f'{self.name}.field -> type')
    QueryValidatorValidator.is_number(self.boost, source=f'{self.name}.boost')
    return True
  
  def execute(self, index, score=True):
    a = index.get_postings(self.field, None)
    if self.boost != 1:
      b = IdsScoresQR(a.ids, np.array(a.scores, dtype=np.float32) * self.boost)
      return b
    else:
      return a
  
  def required_rows(self, index, score=True):
    req_rows = set()
    req_rows.add((index.get_row_key(self.field, None), self.field))
    return req_rows
  
  @classmethod
  def from_json(cls, query, parse_query_fn):
    QueryParserValidator.is_dict(query, source=cls.op_name)

    field = query.get('field')
    boost = query.get('boost')

    QueryParserValidator.is_str(field, source=f'{cls.op_name}.field')
    QueryParserValidator.is_number(boost, source=f'{cls.op_name}.boost')

    return cls(field, boost)



from .query_component import QueryComponent
from ..utils import QueryParserValidator, QueryValidatorValidator

class ScoredTerm(QueryComponent):
  op_name = 'scored_term'
  def __init__(self, field, value):
    super().__init__('scored_term')
    self.field = field
    self.value = value
  
  def validate(self, index, score=True):
    super().validate(index, score=score)
    QueryValidatorValidator.is_str(self.field, source=f'{self.name}.field')
    QueryValidatorValidator.is_str(self.value, source=f'{self.name}.value')
    
    field_type = index.mapping.get_type(self.field)
    QueryValidatorValidator.is_not_none(self.field, source=f'{self.name}.field -> type')
    QueryValidatorValidator.is_valid_type(field_type, {'scored_keyword'}, source=f'{self.name}.field -> type')
    return True

  def execute(self, index, score=True):
    return index.get_postings(self.field, self.value)
  
  def required_rows(self, index, score=True):
    req_rows = set()
    req_rows.add((index.get_row_key(self.field, self.value), self.field))
    return req_rows
  
  @classmethod
  def from_json(cls, query, parse_query_fn):
    QueryValidatorValidator.is_dict(query, source=cls.op_name)
    field = query.get('field')
    value = query.get('value')

    QueryValidatorValidator.is_str(field, source=f'{cls.op_name}.field')
    QueryValidatorValidator.is_str(value, source=f'{cls.op_name}.value')

    return cls(field, value)



from .wrapped import is_sorted, sorted_intersect1d, \
    sorted_in1d, sorted_union1d, sorted_multi_union1d, \
    sorted_sum_scores, sorted_multi_sum_scores, \
    sorted_sum_scores_must, sorted_multi_sum_scores_must, \
    combine_must_should
